/*
 * Copyright (c) 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <complex.h>
#include "pnfft.h"
#include "ipnfft.h"
#include "bessel_i0.h"
#include "bessel_i1.h"
#include "bspline.h"
#include "sinc.h"
#include "matrix_D.h"

#define PNFFT_ENABLE_CALC_INTPOL_NODES 0
#define PNFFT_ENABLE_INTERLACING 0
#define USE_EWALD_SPLITTING_FUNCTION_AS_WINDOW 1

static PNX(plan) mkplan(
    void);
static R interpolation_odd(
    R x, R epsI, int order, INT num_nodes,
    const R *table);
static R interpolation_even(
    R x, R epsI, int order, INT num_nodes,
    const R *table);
static R intpol_even_lin(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);
static R intpol_even_quad(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);
static R intpol_even_kub(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);
static R intpol_odd_lin(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);
static R intpol_odd_quad(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);
static R intpol_odd_kub(
    const R x, const R *table,
    const INT num_nodes, const R eps_I);

static void local_size_B(
    const PNX(plan) ths,
    INT *local_no, INT *local_no_start);
static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above);
static void lowest_summation_index(
    const INT *n, int m, const R *x,
    const INT *local_no_start, const INT *gcells_below,
    INT *u_j);
static void local_array_size(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    INT *local_ngc);

static void pre_psi_tensor_intpol(
    const INT *n, int m, int cutoff, const R *x,
    int intpol_order, INT intpol_num_nodes, R **intpol_tables,
    R *pre_psi);
static void pre_psi_tensor(
    const INT *n, const R *b, int m, int cutoff, const R *x,
    const R *exp_const, R *spline_coeffs,
    unsigned pnfft_flags,
    R *pre_psi);
static void pre_psi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const,
    R *fg_psi);
static void pre_psi_tensor_bspline(
    const INT *n, int m, int cutoff,
    const R *x, R *spline_coeffs, 
    R *pre_psi);
static void pre_psi_tensor_sinc_power(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, 
    R *pre_psi);
static void pre_psi_tensor_bessel_i0(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi);
static void pre_psi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi);

static void pre_dpsi_tensor_intpol(
    const INT *n, int m, int cutoff, const R *x,
    int intpol_order, INT intpol_num_nodes, R **intpol_tables_dpsi,
    R *pre_dpsi);
static void pre_dpsi_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, R *spline_coeffs, const R *pre_psi,
    unsigned pnfft_flags,
    R *pre_dpsi);
static void pre_dpsi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *fg_psi,
    R *fg_dpsi);
static void pre_dpsi_tensor_bspline(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, R *spline_coeffs, const R *pre_psi,
    R *pre_dpsi);
static void pre_dpsi_tensor_sinc_power(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_dpsi);
static void pre_dpsi_tensor_bessel_i0(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_dpsi);
static void pre_dpsi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_dpsi);
static void sort_nodes_for_better_cache_handle(
    int d, const INT *n, int m, INT local_x_num, const R *local_x,
    INT *ar_x);
static void sort_nodes_for_better_cache_handle_qsort_3d(
    const INT *n, int m, INT local_M, const R *x,
    INT *sort);
static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j);
static void get_mpi_cart_dims(
    MPI_Comm comm_cart,
    int *dims);
static int compare_INT(
    const void* a, const void* b);

static R window_bessel_i0_1d(
    R x, INT n, R b, int m);
static R window_bessel_i0_derivative_1d(
    R x, INT n, R b, int m);

static R kaiser_bessel_1d(
    R x, INT n, R b, int m);
static R kaiser_bessel_derivative_1d(
    R x, INT n, R b, int m, R psi);


static void init_intpol_table_psi(
    INT num_nodes, INT n, int m, int dim,
    const PNX(plan) wind_param,
    R *table);
static void init_intpol_table_dpsi(
    INT num_nodes, INT n, int m, int dim,
    const PNX(plan) wind_param,
    R *table);
static R psi_gaussian(
    R x, INT n, R b);
static R dpsi_gaussian(
    R x, INT n, R b);
static R psi_bspline(
    R x, INT n, int m, R *spline_coeffs);
static R dpsi_bspline(
    R x, INT n, int m, R *spline_coeffs);
static R psi_sinc_power(
    R x, INT n, R b, int m);
static R dpsi_sinc_power(
    R x, INT n, R b, int m);
static R psi_bessel_i0(
    R x, INT n, R b, int m);
static R dpsi_bessel_i0(
    R x, INT n, R b, int m);
static R psi_kaiser(
    R x, INT n, R b, int m);
static R dpsi_kaiser(
    R x, INT n, R b, int m);





/* TODO: This function calculates the number of minimum samples of the 2-point-Taylor regularized
 * kernel function 1/x to reach a certain relative error 'eps'. Our windows are likely to be nicer,
 * so this is a good starting point. Think about the optimal number for every window. */
#if PNFFT_ENABLE_CALC_INTPOL_NODES
static INT max_I(INT a, INT b)
{
  return a >= b ? a : b;
}

static INT calc_intpol_num_nodes(
    int intpol_order, R eps
    )
{
  /* use interpolation order 3 as default case */
  switch(intpol_order){
    case 1: return (INT) PNFFT_CEIL(1.7/PNFFT_POW(eps,1.0/2.0));
    case 2: return (INT) PNFFT_CEIL(2.2/PNFFT_POW(eps,1.0/3.0));
    default: return max_I(10, (INT) PNFFT_CEIL(1.4/PNFFT_POW(eps,1.0/4.0)));
  }
}
#endif

static void init_intpol_table_psi(
    INT num_nodes, INT n, int m, int dim,
    const PNX(plan) wind_param,
    R *table
    )
{
  R r = (R)(m+1)/n;

  for(INT k=0; k<num_nodes+3; k++)
    table[k] = PNX(psi)(wind_param, dim, r * (R) k / num_nodes);
}

static void init_intpol_table_dpsi(
    INT num_nodes, INT n, int m, int dim,
    const PNX(plan) wind_param,
    R *table
    )
{
  R r = (R)(m+1)/n;

  for(INT k=0; k<num_nodes+3; k++)
    table[k] = PNX(dpsi)(wind_param, dim, r * (R) k / num_nodes);
}




/* The local borders of the FFT output imply the box of nodes
 * that can be handled by the current process. If x_max < 0.5,
 * the FFT output size 'no' includes some extra elements at the
 * beginning and the end (the number of extra elements depends on
 * the interpolation order 'm').
 * Therefore, we must explicitly check, if the borders are not smaller
 * then -x_max and not bigger then x_max.
 */
void PNX(node_borders)(
    const INT *n,
    const INT* local_no, const INT *local_no_start,
    const R* x_max,
    R *lo, R *up
    )
{
  for(int t=0; t<3; t++){
    lo[t] = local_no_start[t] / ((R) n[t]);    
    up[t] = (local_no_start[t] + local_no[t]) / ((R) n[t]);

//     if(pnfft_flags & PNFFT_SHIFT_OUTPUT){
      /* assure, that borders lie in [-x_max,x_max] */
      if(lo[t] < -x_max[t])
        lo[t] = -x_max[t];
      if(lo[t] > x_max[t])
        lo[t] = x_max[t];
      if(up[t] < -x_max[t])
        up[t] = -x_max[t];
      if(up[t] > x_max[t])
        up[t] = x_max[t];
//     } else {
//       /* assure, that borders lie in [0, x_max] or [1-x_max,1] */
//       if(lo[t] < 0.5){
//         if(lo[t] > x_max[t])
//           lo[t] = x_max[t];
// 
//       if(lo[t] < -x_max[t])
//         lo[t] = -x_max[t];
//       if(lo[t] > x_max[t])
//         lo[t] = x_max[t];
//       if(up[t] < -x_max[t])
//         up[t] = -x_max[t];
//       if(up[t] > x_max[t])
//         up[t] = x_max[t];
//     }
// 
// 
// 
// 
  }
}

static void local_size_B(
    const PNX(plan) ths,
    INT *local_no, INT *local_no_start
    )
{
  for(int t=0; t<ths->d; t++){
    local_no[t] = ths->local_no[t];
    local_no_start[t] = ths->local_no_start[t];
  }
}

void PNX(local_size_internal)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart,
//     MPI_Comm comm_cart, unsigned pnfft_flags,
    INT *local_N, INT *local_N_start,
    INT *local_no, INT *local_no_start
    )
{
  INT alloc_local, howmany = 1;

  alloc_local = PX(local_size_many_dft)(3, n, N, no, howmany,
      PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart, PFFT_TRANSPOSED_IN,
      local_N, local_N_start, local_no, local_no_start);

  /* map input index set from 0:N/2-1 to -N/2:N/2-1 */
//   if(pnfft_flags & PNFFT_SHIFT_INPUT)
    for(int t=0; t<3; t++)
      local_N_start[t] -= N[t]/2;

  /* map output index set from 0:no/2-1 to -no/2:no/2-1 */
//   if(pnfft_flags & PNFFT_SHIFT_OUTPUT)
    for(int t=0; t<3; t++)
      local_no_start[t] -= no[t]/2;
}



/* N - size of NFFT
 * n - oversampled FFT size
 * no - FFT output size (if nodes are only in a subset the array) */
PNX(plan) PNX(init_internal)(
    int d, const INT *N, const INT *n, const INT *no,
    INT local_M, int m,
    unsigned pnfft_flags, unsigned fftw_flags,
    MPI_Comm comm_cart
    )
{
  INT howmany = 1;
  INT alloc_local_data_in, alloc_local_gc, alloc_local_f_hat;
  INT gcells_below[3], gcells_above[3];
  INT local_no[3], local_no_start[3];
  INT local_ngc[3], local_gc_start[3];
  PNX(plan) ths;

  /* TODO: apply some parameter checks */
  
  ths = mkplan();

  ths->d = d;
  ths->m= m;
  ths->local_M = local_M;

  ths->N = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->n = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->no= (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  for(int t=0; t<d; t++){
    ths->N[t]= N[t];
    ths->n[t]= n[t];
    ths->no[t]= no[t];
  }

  ths->local_N        = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->local_N_start  = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->local_no       = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->local_no_start = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);

  ths->pnfft_flags = pnfft_flags;
  ths->fftw_flags = fftw_flags;

  MPI_Comm_dup(comm_cart, &(ths->comm_cart));
  get_mpi_cart_dims(comm_cart, ths->np);
  
  ths->cutoff = 2*m+2;
  ths->N_total = ths->n_total = 1;
  for(int t=0; t<d; t++){
    ths->N_total *= N[t];
    ths->n_total *= n[t];
  }
  /* x_max is filled in init_guru */
  ths->x_max = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  ths->sigma = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  for(int t = 0;t < d; t++)
    ths->sigma[t] = ((R)n[t])/N[t];

  /* Precompute Flags not yet implemented */
  if(pnfft_flags & PNFFT_PRE_PSI){
    PX(printf)(comm_cart, "!!! Warning: Precompute Flag PRE_PSI is not yet implemented in PNFFT !!!\n");
    ths->pnfft_flags &= (~PNFFT_PRE_PSI); /* needed for correct pnfft_finalize */
  }
  if(pnfft_flags & PNFFT_PRE_FULL_PSI){
    PX(printf)(comm_cart, "!!! Warning: Precompute Flag PRE_FULL_PSI is not yet implemented in PNFFT !!!\n");
    ths->pnfft_flags &= (~PNFFT_PRE_FULL_PSI); /* needed for correct pnfft_finalize */
  }
  if(pnfft_flags & PNFFT_PRE_FG_PSI){
    PX(printf)(comm_cart, "!!! Warning: Precompute Flag PRE_FG_PSI is not yet implemented in PNFFT !!!\n");
    ths->pnfft_flags &= (~PNFFT_PRE_FG_PSI); /* needed for correct pnfft_finalize */
  }

  /* init window specific parameters */
  ths->spline_coeffs = NULL; /* needed for correct pnfft_finalize */
  ths->exp_const = NULL; /* needed for correct pnfft_finalize */
  ths->b = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  if(pnfft_flags & PNFFT_WINDOW_GAUSSIAN){
    for(int t=0; t<d; t++)
      ths->b[t]= ((R)m / PNFFT_PI) * K(2.0)*ths->sigma[t] / (K(2.0)*ths->sigma[t]-K(1.0));
#if USE_EWALD_SPLITTING_FUNCTION_AS_WINDOW
    for(int t=0; t<d; t++){
      R C=0.976; /* strange constant from paper by D. Lindbo */
//       R B=10.0; /* box length */
//       R alpha=0.573; /* tuned Ewald splitting parameter */
//       R tmp = (R) ths->n[t] / alpha / B;
//       ths->b[t] -= tmp*tmp;
//       R tmp = (R) ths->n[t];
      ths->b[t] = 2.0 * (R)ths->m / (PNFFT_PI * C*C);
    }
//     for(int t=0; t<d; t++){
//       ths->b[t] = 1.2732;
//     }
#endif
    ths->exp_const = (R*) PNX(malloc)(sizeof(R) * (size_t) d * ths->cutoff);
    for(int t=0; t<d; t++)
      for(int s=0; s<ths->cutoff; s++)
        ths->exp_const[ths->cutoff*t+s] = PNFFT_EXP(-s*s/ths->b[t])/(PNFFT_SQRT(PNFFT_PI*ths->b[t]));
  } else if(pnfft_flags & PNFFT_WINDOW_BSPLINE){
    /* malloc array for scratch values of de Boor algorithm, no need to initialize */
    ths->spline_coeffs= (R*) PNX(malloc)(sizeof(R)*2*ths->m);
  } else if(pnfft_flags & PNFFT_WINDOW_SINC_POWER){
    /* malloc array for scratch values of de Boor algorithm, no need to initialize */
    ths->spline_coeffs= (R*) PNX(malloc)(sizeof(R)*2*ths->m);
    for(int t=0; t<d; t++)
      ths->b[t]= (R)m * (K(2.0)*ths->sigma[t]) / (K(2.0)*ths->sigma[t]-K(1.0));
  } else if(pnfft_flags & PNFFT_WINDOW_BESSEL_I0){
    for(int t=0; t<d; t++)
      ths->b[t] = (R) PNFFT_PI * (K(2.0) - K(1.0)/ths->sigma[t]);
  } else { /* default window function is Kaiser-Bessel */
    for(int t=0; t<d; t++)
      ths->b[t] = (R) PNFFT_PI * (K(2.0) - K(1.0)/ths->sigma[t]);
  }

  get_size_gcells(m, ths->cutoff,
      gcells_below, gcells_above);

  alloc_local_data_in = PX(local_size_many_dft)(3, n, N, no, howmany,
      PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart, PFFT_TRANSPOSED_IN,
      ths->local_N, ths->local_N_start, local_no, local_no_start);

  alloc_local_gc = PX(local_size_many_gc)(3, local_no, local_no_start,
      alloc_local_data_in, howmany, gcells_below, gcells_above,
      local_ngc, local_gc_start);

  /* update offsets such that they run from -N/2,...,N/2-1 */
  PNX(local_size_internal)(N, n, no, comm_cart,
      ths->local_N, ths->local_N_start, ths->local_no, ths->local_no_start);
  
  alloc_local_f_hat = PNX(prod_INT)(3, ths->local_N);
  if(pnfft_flags & PNFFT_MALLOC_F_HAT)
    ths->f_hat = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_f_hat);

  if(pnfft_flags & PNFFT_FFT_INIT){
    ths->g2 = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_gc);
    if(pnfft_flags & PNFFT_FFT_OUT_OF_PLACE)
      ths->g1 = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_data_in);
    else
      ths->g1 = ths->g2;
    
    ths->pfft_forw = PX(plan_many_dft)(3, n, N, no, howmany,
        PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, ths->g1, ths->g2, comm_cart,
        FFTW_FORWARD, PFFT_TRANSPOSED_IN, fftw_flags);
    
    ths->pfft_back = PX(plan_many_dft)(3, n, no, N, howmany,
        PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, ths->g2, ths->g1, comm_cart,
        FFTW_BACKWARD, PFFT_TRANSPOSED_OUT, fftw_flags);

    ths->gcplan = PX(plan_many_cgc)(3, no, howmany, PFFT_DEFAULT_BLOCKS,
        gcells_below, gcells_above, ths->g2, comm_cart, 0);
  } else {
    ths->g1 = ths->g2 = NULL;
    ths->pfft_forw = ths->pfft_back = NULL;
    ths->gcplan = NULL;
  }

  if(pnfft_flags & PNFFT_PRE_PHI_HAT){
    ths->pre_inv_phi_hat_trafo = (C*) malloc(sizeof(C)*(ths->local_N[0] + ths->local_N[1] + ths->local_N[2]));
    ths->pre_inv_phi_hat_adj   = (C*) malloc(sizeof(C)*(ths->local_N[0] + ths->local_N[1] + ths->local_N[2]));
    
    PNX(precompute_inv_phi_hat_trafo)(ths,
        ths->pre_inv_phi_hat_trafo);
    PNX(precompute_inv_phi_hat_adj)(ths,
        ths->pre_inv_phi_hat_adj);
  } else {
    ths->pre_inv_phi_hat_trafo = NULL; /* needed for correct pnfft_finalize */
    ths->pre_inv_phi_hat_adj   = NULL; /* needed for correct pnfft_finalize */
  }

  if(pnfft_flags & PNFFT_PRE_LIN_PSI)
    ths->intpol_order = 1;
  else if(pnfft_flags & PNFFT_PRE_QUAD_PSI)
    ths->intpol_order = 2;
  else if(pnfft_flags & PNFFT_PRE_KUB_PSI)
    ths->intpol_order = 3;
  else
    ths->intpol_order = -1;

  if(pnfft_flags & PNFFT_PRE_INTPOL_PSI){
#if PNFFT_ENABLE_CALC_INTPOL_NODES
    ths->intpol_num_nodes = calc_intpol_num_nodes(ths->intpol_order, 1e-16);
#else
    ths->intpol_num_nodes = 70000;
#endif
    ths->intpol_tables_psi  = (R**) PNX(malloc)(sizeof(R*) * (size_t) d);
    ths->intpol_tables_dpsi = (R**) PNX(malloc)(sizeof(R*) * (size_t) d);
    for(int t=0; t<d; t++){
      ths->intpol_tables_psi[t] = (R*) PNX(malloc)(sizeof(R)*(ths->intpol_num_nodes+3));
      init_intpol_table_psi(ths->intpol_num_nodes, ths->n[t], ths->m, t, ths,
          ths->intpol_tables_psi[t]);
      ths->intpol_tables_dpsi[t] = (R*) PNX(malloc)(sizeof(R)*(ths->intpol_num_nodes+3));
      init_intpol_table_dpsi(ths->intpol_num_nodes, ths->n[t], ths->m, t, ths,
          ths->intpol_tables_dpsi[t]);
    }
  }

  /* init compute flags before malloc f and f_grad */
  ths->compute_flags = 0;

  /* allocate mem and adjust pnfft_flags, compute_flags */
  PNX(malloc_x)(ths, pnfft_flags);
  PNX(malloc_f)(ths, pnfft_flags);
  PNX(malloc_grad_f)(ths, pnfft_flags);

  return ths;
}


static PNX(plan) mkplan(
    void
    )
{
  PNX(plan) ths = (plan_s*) malloc(sizeof(plan_s));

  ths->f_hat  = NULL;
  ths->f      = NULL;
  ths->grad_f = NULL;
  ths->x      = NULL;
  ths->N      = NULL;
  ths->sigma  = NULL;
  ths->n      = NULL;
  ths->no     = NULL;
  ths->x_max  = NULL;

  ths->local_N        = NULL;
  ths->local_N_start  = NULL;
  ths->local_no       = NULL;
  ths->local_no_start = NULL;

  ths->b              = NULL;
  ths->exp_const      = NULL;
  ths->spline_coeffs  = NULL;

  ths->pre_inv_phi_hat_trafo  = NULL;
  ths->pre_inv_phi_hat_adj    = NULL;

  ths->g1 = NULL;
  ths->g2 = NULL;
  
  ths->intpol_tables_psi  = NULL;
  ths->intpol_tables_dpsi = NULL;

  ths->timer_trafo = PNX(mktimer)();
  ths->timer_adj   = PNX(mktimer)();

  return ths;
}


void PNX(rmplan)(
    PNX(plan) ths
    )
{
  /* plan was already destroyed or never initialized */
  if(ths==NULL)
    return;

  PNX(rmtimer)(ths->timer_trafo);
  PNX(rmtimer)(ths->timer_adj);

  /* free memory */
  free(ths);
  /* ths=NULL; would be senseless, since we can not change the pointer itself */
}





void PNX(malloc_x)(
    PNX(plan) ths, unsigned pnfft_flags
    )
{
  ths->x = NULL;
  if( (pnfft_flags & PNFFT_MALLOC_X) && (ths->local_M > 0) ){
    ths->x = (R*) PNX(malloc)(sizeof(R) * (size_t) ths->d*ths->local_M);
    ths->pnfft_flags |= PNFFT_MALLOC_X;
  }
}

void PNX(free_x)(
    PNX(plan) ths, unsigned pnfft_finalize_flags
    )
{
  if( (pnfft_finalize_flags & PNFFT_FREE_X) && (ths->x != NULL) ){
    PNX(free)(ths->x);
    ths->pnfft_flags &= ~PNFFT_MALLOC_X;
  }
}

void PNX(malloc_f)(
    PNX(plan) ths, unsigned pnfft_flags
    )
{
  ths->f = NULL;
  if( (pnfft_flags & PNFFT_MALLOC_F) && (ths->local_M > 0) ){
    ths->f = (C*) PNX(malloc)(sizeof(C) * (size_t) ths->local_M);
    ths->pnfft_flags |= PNFFT_COMPUTE_F;
    ths->compute_flags |= PNFFT_COMPUTE_F;
  }
}

void PNX(free_f)(
    PNX(plan) ths, unsigned pnfft_finalize_flags
    )
{
  if( (pnfft_finalize_flags & PNFFT_FREE_F) && (ths->f != NULL) ){
    PNX(free)(ths->f);
    ths->pnfft_flags &= ~PNFFT_MALLOC_F;
    ths->compute_flags &= ~PNFFT_COMPUTE_F;
  }
}

void PNX(malloc_grad_f)(
    PNX(plan) ths, unsigned pnfft_flags
    )
{
  ths->grad_f = NULL;
  if( (pnfft_flags & PNFFT_MALLOC_GRAD_F) && (ths->local_M > 0) ){
    ths->grad_f = (C*) PNX(malloc)(sizeof(C) * (size_t) ths->d*ths->local_M);
    ths->pnfft_flags |= PNFFT_MALLOC_GRAD_F;
    ths->compute_flags |= PNFFT_COMPUTE_GRAD_F;
  }
}

void PNX(free_grad_f)(
    PNX(plan) ths, unsigned pnfft_finalize_flags
    )
{
  if( (pnfft_finalize_flags & PNFFT_FREE_GRAD_F) && (ths->grad_f != NULL) ){
    PNX(free)(ths->grad_f);
    ths->pnfft_flags &= ~PNFFT_MALLOC_GRAD_F;
    ths->compute_flags &= ~PNFFT_COMPUTE_GRAD_F;
  }
}


void PNX(trafo_F)(
    PNX(plan) ths
    )
{
  PX(execute)(ths->pfft_forw);
}

void PNX(adjoint_F)(
    PNX(plan) ths
    )
{
  PX(execute)(ths->pfft_back);
}


/* make use of tensor structure, precompute exp(...) for inner loops */
static void fftshift_N_half(
    const INT *N, const INT *n, const INT *no,
    const INT *local_no, const INT *local_no_start,
    int sign,
    C *g
    )
{
  INT l0, l1, l2, m=0;

//   double timer = -MPI_Wtime();

  /* plain loop:
   * requires n^3 evaluations of exp(), 3*n^3 multiplications */
//   for(l0=local_no_start[0]; l0<local_no[0]+local_no_start[0]; l0++)
//     for(l1=local_no_start[1]; l1<local_no[1]+local_no_start[1]; l1++)
//       for(l2=local_no_start[2]; l2<local_no[2]+local_no_start[2]; l2++, m++)
//         g[m] *= PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[0]*l0/((R)n[0]) + N[1]*l1/((R) n[1]) + N[2]*l2/((R) n[2]) ));
 
  /* precompute exp():
   * requires 3*n evaluations of exp(), 3*n^3 multiplications, 3*n complex numbers in memory */
//   C *exp0, *exp1, *exp2;
//   exp0 = (C*) malloc(sizeof(C)*local_no[0]);
//   exp1 = (C*) malloc(sizeof(C)*local_no[1]);
//   exp2 = (C*) malloc(sizeof(C)*local_no[2]);
//   
//   for(l0=local_no_start[0], m=0; l0<local_no[0]+local_no_start[0]; l0++, m++)
//     exp0[m] = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[0]*l0/((R)n[0]) ));
//   for(l1=local_no_start[1], m=0; l1<local_no[1]+local_no_start[1]; l1++, m++)
//     exp1[m] = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[1]*l1/((R)n[1]) ));
//    for(l2=local_no_start[2], m=0; l2<local_no[2]+local_no_start[2]; l2++, m++)
//     exp2[m] = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[2]*l2/((R)n[2]) ));
// 
//   m=0;
//   for(l0=0; l0<local_no[0]; l0++)
//     for(l1=0; l1<local_no[1]; l1++)
//       for(l2=0; l2<local_no[2]; l2++, m++)
//         g[m] *= exp0[l0] * exp1[l1] * exp2[l2];
// 
//   free(exp0); free(exp1); free(exp2);

  /* precompute exp() and premultiply:
   * requires 3*n evaluations of exp(), 2*n^3 multiplications, 2*n complex numbers in memory */
  C exp0, exp01, *pre_exp1, *pre_exp2;

  pre_exp1 = (C*) malloc(sizeof(C)*local_no[1]);
  pre_exp2 = (C*) malloc(sizeof(C)*local_no[2]);

  for(l1=local_no_start[1], m=0; l1<local_no[1]+local_no_start[1]; l1++, m++)
    pre_exp1[m] = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[1]*l1/((R)n[1]) ));
  for(l2=local_no_start[2], m=0; l2<local_no[2]+local_no_start[2]; l2++, m++)
    pre_exp2[m] = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[2]*l2/((R)n[2]) ));

  m=0;
  for(l0=local_no_start[0]; l0<local_no[0]+local_no_start[0]; l0++){
    exp0 = PNFFT_CEXP(-sign * PNFFT_PI * _Complex_I * ( N[0]*l0/((R)n[0]) ));
    for(l1=0; l1<local_no[1]; l1++){
      exp01 = exp0 * pre_exp1[l1];
      for(l2=0; l2<local_no[2]; l2++, m++){
        g[m] *= exp01 * pre_exp2[l2];
      }
    }
  }

  free(pre_exp1); free(pre_exp2);
  
//   timer += MPI_Wtime();
//   printf("!!! time for fftshift_N_half: %e\n", timer);
}


void PNX(trafo_B)(
    PNX(plan) ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3];
  R *pre_psi = NULL;
  R *pre_dpsi = NULL;
 
  local_size_B(ths,
      local_no, local_no_start);

#if PNFFT_ENABLE_DEBUG
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  R rsum, grsum;
  R rsum_derive, grsum_derive;
  C csum, gcsum;
  
  if(!myrank) fprintf(stderr, "no = [%td, %td, %td]\n", ths->no[0], ths->no[1], ths->no[2]);
//  if(!myrank) fprintf(stderr, "local_no = [%td, %td, %td], local_ngc = [%td, %td, %td], local_ngc_total = %td\n",
//      local_no[0], local_no[1], local_no[2], local_ngc[0], local_ngc[1], local_ngc[2], local_ngc_total);
      
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients before twiddles: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif

  /* perform fftshift */
  ths->timer_trafo[PNFFT_TIMER_SHIFT_INPUT] -= MPI_Wtime();
//   if(ths->pnfft_flags & PNFFT_SHIFT_INPUT)
    fftshift_N_half(ths->N, ths->n, ths->no, local_no, local_no_start, FFTW_FORWARD,
        ths->g2);
  ths->timer_trafo[PNFFT_TIMER_SHIFT_INPUT] += MPI_Wtime();

  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients before ghostcell send: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif

  /* send ghost cells in ring */
  ths->timer_trafo[PNFFT_TIMER_GCELLS] -= MPI_Wtime();
  PX(exchange)(ths->gcplan);
  ths->timer_trafo[PNFFT_TIMER_GCELLS] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  INT local_ngc_total = PNX(prod_INT)(3, local_ngc);
#endif

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_ngc_total; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients after ghostcell send: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif  

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(INT j=0; j<3*ths->local_M; j++)
    rsum += fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of x before sort: %e\n", grsum);
#endif

  /* sort indices for better cache handling */
  if(ths->pnfft_flags & PNFFT_SORT_NODES){
    ths->timer_trafo[PNFFT_TIMER_SORT_NODES] -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->d, ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_trafo[PNFFT_TIMER_SORT_NODES] += MPI_Wtime();
  }

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(j=0; j<3*ths->local_M; j++)
    rsum += fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of x after sort: %e\n", grsum);
#endif

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    rsum_derive = 0.0;
#endif

  ths->timer_trafo[PNFFT_TIMER_LOOP_B] -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    pre_dpsi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->pnfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;
    
    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);
  
#if PNFFT_ENABLE_DEBUG
    for(int t=0; t<3*cutoff; t++)
      rsum += fabs(pre_psi[t]);
#endif
  
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
        pre_dpsi_tensor_intpol(
            ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
            ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_dpsi,
            pre_dpsi);
      else
        pre_dpsi_tensor(
            ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->spline_coeffs,
            pre_psi, ths->pnfft_flags,
            pre_dpsi);
    }

#if PNFFT_ENABLE_DEBUG
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
      for(int t=0; t<3*cutoff; t++)
        rsum_derive += fabs(pre_dpsi[t]);
#endif

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);
    
    /* compute f */
    if(ths->compute_flags & PNFFT_COMPUTE_F){
      ths->f[j] = 0;
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++)
#if PNFFT_ENABLE_INTERLACING
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * creal(ths->g2[m2]);
#else
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
#endif
    }

    /* compute grad_f */
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      for(int t=0; t<ths->d; t++)
        ths->grad_f[ths->d*j+t] = 0;
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++){
#if PNFFT_ENABLE_INTERLACING
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * creal(ths->g2[m2]);
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * creal(ths->g2[m2]);
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * creal(ths->g2[m2]);
#else
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * ths->g2[m2];
#endif
          }
        }
      }
    }

#if PNFFT_ENABLE_INTERLACING
    R x[3];
    
    /* shift x by half the mesh width */
    for(int t=0; t<3; t++)
      x[t] = ths->x[ths->d*j+t] + 0.5/ths->n[t];

    /* We need to compute the lowest summation index before we fold x back into [-0.5,0.5).
     * Otherwise u_j may be also folded and gets less than the local offset local_no_start. */
    lowest_summation_index(
        ths->n, ths->m, x, local_no_start, gcells_below,
        u_j);
    
    /* assure -0.5 <= x < 0.5 */
    for(int t=0; t<3; t++)
      if(x[t] >= 0.5)
        x[t] -= 1.0;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, x,
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, x,
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
        pre_dpsi_tensor_intpol(
            ths->n, ths->m, ths->cutoff, x,
            ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_dpsi,
            pre_dpsi);
      else
        pre_dpsi_tensor(
            ths->n, ths->b, ths->m, ths->cutoff, x, ths->spline_coeffs,
            pre_psi, ths->pnfft_flags,
            pre_dpsi);
    }
    
    /* compute f with shifted x */
    if(ths->compute_flags & PNFFT_COMPUTE_F){
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++)
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * cimag(ths->g2[m2]);

      ths->f[j] /= 2.0;
    }

    /* compute grad_f with shifted x */
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++){
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * cimag(ths->g2[m2]);
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * cimag(ths->g2[m2]);
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * cimag(ths->g2[m2]);
          }
        }
      }
      for(int t=0; t<ths->d; t++)
        ths->grad_f[ths->d*j+t] /= 2.0;
    }
#endif
  }

  ths->timer_trafo[PNFFT_TIMER_LOOP_B] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of pre_psi: %e\n", grsum);
#endif  

#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    MPI_Reduce(&rsum_derive, &grsum_derive, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of pre_dpsi: %e\n", grsum_derive);
  }
#endif  

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(j=0; j<ths->local_M; j++)
    csum += fabs(creal(ths->f[j])) + _Complex_I * fabs(cimag(ths->f[j]));
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of f: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += fabs(creal(ths->grad_f[3*j+0])) + _Complex_I * fabs(cimag(ths->grad_f[3*j+0]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 1st component of grad_f: %e + I* %e\n", creal(gcsum), cimag(gcsum));
  }
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += fabs(creal(ths->grad_f[3*j+1])) + _Complex_I * fabs(cimag(ths->grad_f[3*j+1]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 2nd component of grad_f: %e + I* %e\n", creal(gcsum), cimag(gcsum));
  }
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += fabs(creal(ths->grad_f[3*j+2])) + _Complex_I * fabs(cimag(ths->grad_f[3*j+2]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 3rd component of grad_f: %e + I* %e\n", creal(gcsum), cimag(gcsum));
  }
#endif
  
  if(ths->pnfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);

  PNX(free)(pre_psi);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    PNX(free)(pre_dpsi);
}


void PNX(adjoint_B)(
    PNX(plan) ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3], local_ngc_total;
  R *pre_psi = NULL;
  
  local_size_B(ths,
      local_no, local_no_start);

  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

  local_ngc_total = PNX(prod_INT)(3, local_ngc);
  memset(ths->g2, 0, local_ngc_total * sizeof(C));

#if PNFFT_ENABLE_DEBUG
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  R rsum, grsum;
  C csum, gcsum;
  
  if(!myrank) fprintf(stderr, "no = [%td, %td, %td]\n", ths->no[0], ths->no[1], ths->no[2]);
//  if(!myrank) fprintf(stderr, "local_no = [%td, %td, %td], local_ngc = [%td, %td, %td], local_ngc_total = %td\n",
//      local_no[0], local_no[1], local_no[2], local_ngc[0], local_ngc[1], local_ngc[2], local_ngc_total);
      
  rsum = 0.0;
  for(INT j=0; j<3*ths->local_M; j++)
    rsum += fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of x before sort: %e\n", grsum);
#endif

  /* sort indices for better cache handling */
  if(ths->pnfft_flags & PNFFT_SORT_NODES){
    ths->timer_adj[PNFFT_TIMER_SORT_NODES] -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->d, ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_adj[PNFFT_TIMER_SORT_NODES] += MPI_Wtime();
  }
  
#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(j=0; j<3*ths->local_M; j++)
    rsum += fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of x after sort: %e\n", grsum);
#endif
  
#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(j=0; j<ths->local_M; j++)
    csum += fabs(creal(ths->f[j])) + _Complex_I * fabs(cimag(ths->f[j]));
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of f: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif
  
#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
#endif

  ths->timer_adj[PNFFT_TIMER_LOOP_B] -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->pnfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

#if PNFFT_ENABLE_DEBUG
    for(int t=0; t<3*cutoff; t++)
      rsum += fabs(pre_psi[t]);
#endif

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);
    
#if PNFFT_ENABLE_DEBUG
    R lo[3], up[3];
    PNX(node_borders)(ths->n, local_no, local_no_start, ths->x_max,
        lo, up);

    int failed=0;
    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ ){
           if((m2 < 0) && (!failed)){
             fprintf(stderr, "!!! Error: myrank = %d, write to g2 out of range (m2<0), m2=%td !!!\n", myrank, m2);
             failed = 1;
           }
           if((m2 >= local_ngc_total) && (!failed)){
             fprintf(stderr, "!!! Error: write to g2 out of range (m2>=local_ngc_total)i, m2=%td, local_ngc_total=%td !!!\n", m2, local_ngc_total);
             failed = 1;
           }
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->f[j];
        }
      }
    }
    if(failed){
      fprintf(stderr, "myrank = %d, u_j=[%td,%td,%td], m0 = %td !!!\n", myrank, u_j[0], u_j[1], u_j[2], m0);
      fprintf(stderr, "local_M = %td, local_no = [%td, %td, %td]\n",
      ths->local_M, local_no[0], local_no[1], local_no[2]);
      fprintf(stderr, "x = [%f, %f, %f]i, lower_border = [%f, %f, %f], upper_border = [%f, %f, %f]\n",
      ths->x[ths->d*j], ths->x[ths->d*j+1], ths->x[ths->d*j+2],
        lo[0], lo[1], lo[2], up[0], up[1], up[2]);
    }
#else
    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ )
#if PNFFT_ENABLE_INTERLACING
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * creal(ths->f[j]);
#else
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->f[j];
#endif
#endif
#if PNFFT_ENABLE_INTERLACING
    R x[3];

    /* shift x by half the mesh width */
    for(int t=0; t<3; t++)
      x[t] = ths->x[ths->d*j+t] + 0.5/ths->n[t];

    /* We need to compute the lowest summation index before we fold x back into [-0.5,0.5).
     * Otherwise u_j may be also folded and gets less than the local offset local_no_start. */
    lowest_summation_index(
        ths->n, ths->m, x, local_no_start, gcells_below,
        u_j);

    /* assure -0.5 <= x < 0.5 */
    for(int t=0; t<3; t++)
      if(x[t] >= 0.5)
        x[t] -= 1.0;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, x,
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, x,
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ )
           ths->g2[m2] += I * pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * creal(ths->f[j]);
#endif
  }
  ths->timer_adj[PNFFT_TIMER_LOOP_B] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of pre_psi: %e\n", grsum);
#endif  

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_ngc_total; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients before ghostcell reduce: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif  

  /* reduce ghost cells in ring */
  ths->timer_adj[PNFFT_TIMER_GCELLS] -= MPI_Wtime();
  PX(reduce)(ths->gcplan);
  ths->timer_adj[PNFFT_TIMER_GCELLS] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients after ghostcell reduce: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif
 
  /* perform fftshift */
  ths->timer_adj[PNFFT_TIMER_SHIFT_INPUT] -= MPI_Wtime();
//   if(ths->pnfft_flags & PNFFT_SHIFT_INPUT)
    fftshift_N_half(ths->N, ths->n, ths->no, local_no, local_no_start, FFTW_BACKWARD,
        ths->g2);
  ths->timer_adj[PNFFT_TIMER_SHIFT_INPUT] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += fabs(creal(ths->g2[t])) + _Complex_I * fabs(cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients after twiddles: %e + I* %e\n", creal(gcsum), cimag(gcsum));
#endif
  
  if(ths->pnfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);
  PNX(free)(pre_psi);
}


/* Implement ghostcell send for all dimensions */
static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above
    )
{
  for(int t=0; t<3; t++){
    gcells_below[t] = m;
    gcells_above[t] = cutoff - gcells_below[t] - 1;
#if PNFFT_ENABLE_INTERLACING
    gcells_above[t] += 1;
#endif
  }
}

static void lowest_summation_index(
    const INT *n, int m, const R *x,
    const INT *local_no_start, const INT *gcells_below,
    INT *u_j
    )
{
  project_node_to_grid(n, m, x, u_j);
  for(int t=0; t<3; t++)
    u_j[t] = u_j[t] - local_no_start[t] + gcells_below[t];
}


static void local_array_size(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    INT *local_ngc
    )
{
  for(int t=0; t<3; t++)
    local_ngc[t] = local_n[t] + gcells_below[t] + gcells_above[t];
}


static void pre_psi_tensor_intpol(
    const INT *n, int m, int cutoff, const R *x,
    int intpol_order, INT intpol_num_nodes, R **intpol_tables_psi,
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m; 
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] = interpolation_even( (u_j+s)/n[t], (R)(m+1)/n[t],
          intpol_order, intpol_num_nodes, intpol_tables_psi[t]);
  }
}

static void pre_dpsi_tensor_intpol(
    const INT *n, int m, int cutoff, const R *x,
    int intpol_order, INT intpol_num_nodes, R **intpol_tables_dpsi,
    R *pre_dpsi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m; 
    for(int s=0; s<cutoff; s++)
      pre_dpsi[cutoff*t+s] = interpolation_odd( (u_j+s)/n[t], (R)(m+1)/n[t],
          intpol_order, intpol_num_nodes, intpol_tables_dpsi[t]);
  }
}


/* calculate window function */
static void pre_psi_tensor(
    const INT *n, const R *b, int m, int cutoff, const R *x,
    const R *exp_const, R *spline_coeffs,
    unsigned pnfft_flags,
    R *pre_psi
    )
{
  if(pnfft_flags & PNFFT_WINDOW_GAUSSIAN)
    pre_psi_tensor_gaussian(
        n, b, m, cutoff, x, exp_const,
        pre_psi);
  else if(pnfft_flags & PNFFT_WINDOW_BSPLINE)
    pre_psi_tensor_bspline(
        n, m, cutoff, x, spline_coeffs,
        pre_psi);
  else if(pnfft_flags & PNFFT_WINDOW_SINC_POWER)
    pre_psi_tensor_sinc_power(
        n, b, m, cutoff, x,
        pre_psi);
  else if(pnfft_flags & PNFFT_WINDOW_BESSEL_I0)
    pre_psi_tensor_bessel_i0(
        n, b, m, cutoff, x,
        pre_psi);
  else
    pre_psi_tensor_kaiser_bessel(
        n, b, m, cutoff, x,
        pre_psi);
}

static void pre_psi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const,
    R *fg_psi
    )
{
  const int d=3;
  R u_j, exp_sqr, exp_lin, tmp;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - m;
    exp_sqr = PNFFT_EXP( -PNFFT_SQR( n[t]*x[t]-u_j ) / b[t] );
    exp_lin = PNFFT_EXP( 2*( n[t]*x[t]-u_j ) / b[t] );

    tmp = exp_sqr;
    for(int s=0; s<cutoff; s++){
      fg_psi[cutoff*t+s] = tmp * exp_const[cutoff*t+s];
      tmp *= exp_lin;
    }
  }
}

static void pre_psi_tensor_bspline(
    const INT *n, int m, int cutoff,
    const R *x, R *spline_coeffs, 
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    /* Bspline is shifted by m */
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t]; 
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] = PNX(bspline)(
          2*m, u_j + (R)s, spline_coeffs);
  }
}

/* The factor n of the window cancels with the factor 1/n from matrix D. */
static void pre_psi_tensor_sinc_power(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, 
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m; 
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] =
        PNFFT_POW(
            PNX(sinc)( PNFFT_PI * (u_j + s) / b[t]),
            K(2.0)*(R)m
        ) / b[t];
  }
}

static void pre_psi_tensor_bessel_i0(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m;
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] = window_bessel_i0_1d(
          (u_j + s) / n[t], n[t], b[t], m);
  }
}

static void pre_psi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m;
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] = kaiser_bessel_1d(
          (u_j + s) / n[t], n[t], b[t], m);
  }
}


/* calculate window derivative */
static void pre_dpsi_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, R *spline_coeffs, const R *pre_psi,
    unsigned pnfft_flags,
    R *pre_dpsi
    )
{
  if(pnfft_flags & PNFFT_WINDOW_GAUSSIAN)
    pre_dpsi_tensor_gaussian(
        n, b, m, cutoff, x, pre_psi,
        pre_dpsi);
  else if(pnfft_flags & PNFFT_WINDOW_BSPLINE)
    pre_dpsi_tensor_bspline(
        n, b, m, cutoff, x, spline_coeffs, pre_psi,
        pre_dpsi);
  else if(pnfft_flags & PNFFT_WINDOW_SINC_POWER)
    pre_dpsi_tensor_sinc_power(
        n, b, m, cutoff, x, pre_psi,
        pre_dpsi);
  else if(pnfft_flags & PNFFT_WINDOW_BESSEL_I0)
    pre_dpsi_tensor_bessel_i0(
        n, b, m, cutoff, x,
        pre_dpsi);
  else
    pre_dpsi_tensor_kaiser_bessel(
        n, b, m, cutoff, x, pre_psi,
        pre_dpsi);
}

static void pre_dpsi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *fg_psi,
    R *fg_dpsi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = n[t]*x[t] - PNFFT_FLOOR( n[t]*x[t] ) + m;
    for(int s=0; s<cutoff; s++)
      fg_dpsi[cutoff*t+s] = -2.0*n[t]/b[t] * (u_j-s) * fg_psi[cutoff*t+s];
  }
}

static void pre_dpsi_tensor_bspline(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, R *spline_coeffs, const R *pre_psi,
    R *pre_dpsi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    /* Bspline is shifted by m */
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t];
    for(int s=0; s<cutoff; s++)
      pre_dpsi[cutoff*t+s] = (R)n[t] *
        ( PNX(bspline)(2*m-1, u_j + (R)s - 1, spline_coeffs)
          - PNX(bspline)(2*m-1, u_j + (R)s, spline_coeffs) );
  }
}


/* The factor n of the window cancels with the factor 1/n from matrix D. */
static void pre_dpsi_tensor_sinc_power(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_dpsi
    )
{
  const int d=3;
  R u_j, y;

  for(int t=0; t<d; t++){
    u_j =  PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m;
    for(int s=0; s<cutoff; s++){
      y =  PNFFT_PI * (u_j + s) / b[t];
      if(PNFFT_FABS(y) > PNFFT_EPSILON)
        pre_dpsi[cutoff*t+s] =
          2.0 * (R)m * PNFFT_PI * (R)n[t] / b[t] * ( 1.0/y - 1.0/PNFFT_TAN(y) ) * pre_psi[cutoff*t+s];
      else
        pre_dpsi[cutoff*t+s] = K(0.0);
    }
  }
}

static void pre_dpsi_tensor_bessel_i0(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_dpsi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m;
    for(int s=0; s<cutoff; s++)
      pre_dpsi[cutoff*t+s] = window_bessel_i0_derivative_1d(
          (u_j + s) / n[t], n[t], b[t], m);
  }
}


static void pre_dpsi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_dpsi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j =  PNFFT_FLOOR( n[t]*x[t] ) - n[t]*x[t] - m;
    for(int s=0; s<cutoff; s++)
      pre_dpsi[cutoff*t+s] = kaiser_bessel_derivative_1d(
          (u_j + s) / n[t],
          n[t], b[t], m, pre_psi[cutoff*t+s]);
  }
}


static void sort_nodes_for_better_cache_handle_qsort_3d(
    const INT *n, int m, INT local_M, const R *x,
    INT *sort
    )
{ /* sort must be of length 2*local_M */
  int d = 3;
  INT u_j[3];

  if(local_M == 0)
    return;

  for(INT k = 0; k < local_M; k++){
    project_node_to_grid(n, m, x+d*k, u_j);
    sort[2*k]   = PNFFT_PLAIN_INDEX_3D(u_j, n);
    sort[2*k+1] = k;
  }
  qsort(sort, (size_t) local_M, 2*sizeof(INT), compare_INT);
}





/**
 * Sort nodes (index) to get better cache utilization during multiplication
 * with matrix B.
 * The resulting index set is written to ar[2*j+1], the nodes array remains
 * unchanged.
 *
 * \arg n FFTW length (number of oversampled in each dimension)
 * \arg m window length
 * \arg local_x_num number of nodes
 * \arg local_x nodes array
 * \arg ar_x resulting index array
 *
 * \author Toni Volkmer
 */
static void sort_nodes_for_better_cache_handle(
    int d, const INT *n, int m, INT local_x_num, const R *local_x,
    INT *ar_x
    )
{
#if SORT_RADIX
  INT u_j[d], i, j, help, rhigh;
  INT *ar_x_temp;
  INT nprod;

  for(i = 0; i < local_x_num; i++) {
    ar_x[2*i] = 0;
    ar_x[2*i+1] = i;
    for(j = 0; j < d; j++) {
      help = floor( n[j]*local_x[d*i+j] - m);
      u_j[j] = (help%n[j]+n[j])%n[j];

      ar_x[2*i] += u_j[j];
      if (j+1 < d)
        ar_x[2*i] *= n[j+1];
    }
  }

  for (j = 0, nprod = 1; j < d; j++)
    nprod *= n[j];

  rhigh = ceil(log2(nprod)) - 1;

  ar_x_temp = (INT*) PNX(malloc)(2*local_x_num*sizeof(INT));
  PNX(sort_node_indices_radix_lsdf)(local_x_num, ar_x, ar_x_temp, rhigh);
#  ifdef OMP_ASSERT
  for (i = 1; i < local_x_num; i++)
    assert(ar_x[2*(i-1)] <= ar_x[2*i]);
#  endif
  PNX(free)(ar_x_temp);
#else
  sort_nodes_for_better_cache_handle_qsort_3d(n, m, local_x_num, local_x, ar_x);
#endif
}





static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j
    )
{
  for(int t=0; t<3; t++)
    u_j[t] = PNFFT_LRINT(PNFFT_FLOOR(n[t]*x[t])) - m;
}

static void get_mpi_cart_dims(
    MPI_Comm comm_cart,
    int *dims
    )
{
  int maxdims, *periods, *coords;
 
  MPI_Cartdim_get(comm_cart, &maxdims);

  periods = (int*) malloc(sizeof(int) * (size_t) maxdims);
  coords  = (int*) malloc(sizeof(int) * (size_t) maxdims);
  MPI_Cart_get(comm_cart, maxdims, dims, periods, coords);
  free(periods);
  free(coords);
 
  if(maxdims == 2)
    dims[2] = 1;
}

static int compare_INT(
    const void* a, const void* b
    )
{
  return (*(const INT*)a <= *(const INT*)b) ? -1 : 1;
}

/* compact support in real space */
static R window_bessel_i0_1d(
    R x, INT n, R b, int m
    )
{
  /* TODO: try to avoid case d<0, since in theory d >= 0 */
  R d = PNFFT_SQR( (R)m ) - PNFFT_SQR( x*n );
  R r = (d<0) ? PNFFT_SQRT(-d) : PNFFT_SQRT(d);

  /* Compact support in real space */
  return (d>0) ? 0.5 * PNX(bessel_i0)(b*r) : 0.0;
}

static R window_bessel_i0_derivative_1d(
    R x, INT n, R b, int m
    )
{
  R d = PNFFT_SQR( (R)m ) - PNFFT_SQR( x*n );
  R r = (d<0) ? PNFFT_SQRT(-d) : PNFFT_SQRT(d);

  /* Compact support in real space */
  return (d>0) ? 0.5 * b * (R)n * (R)n * x * PNX(bessel_i1)(b*r) / r : 0.0;
}

static R kaiser_bessel_1d(
    R x, INT n, R b, int m
    )
{
  /* TODO: try to avoid case d<0, since in theory d >= 0 */
  R d, r;

  d = PNFFT_SQR( (R)m ) - PNFFT_SQR( x*n );
  r = (d<0) ? PNFFT_SQRT(-d) : PNFFT_SQRT(d);

  if(d < 0)
    return PNFFT_SIN(b*r) / ((R)PNFFT_PI*r);

  /* regular case d >= 0 */
  return (d>0) ? PNFFT_SINH(b*r) / ((R)PNFFT_PI*r) : b/PNFFT_PI;
}

static R kaiser_bessel_derivative_1d(
    R x, INT n, R b, int m, R psi
    )
{
  /* TODO: try to avoid case d<0, since in theory d >= 0 */
  R d, r;

  d = PNFFT_SQR( (R)m ) - PNFFT_SQR( x*n );
  r = (d<0) ? PNFFT_SQRT(-d) : PNFFT_SQRT(d);

  if(d < 0) /* use of -d results in change of sign within the brackets */
    return n*n*x/d * (b*PNFFT_COS(b*r)/PNFFT_PI - psi);

  /* regular case d >= 0 */
  return (d>0) ? n*n*x/d * (b*PNFFT_COSH(b*r)/PNFFT_PI - psi) : -n*m*b*b*b/(3.0*PNFFT_PI);
}




R PNX(psi)(
    const PNX(plan) ths, int dim, R x
    )
{
  if(ths->pnfft_flags & PNFFT_WINDOW_GAUSSIAN)
    return psi_gaussian(x, ths->n[dim], ths->b[dim]);
  else if(ths->pnfft_flags & PNFFT_WINDOW_BSPLINE)
    return psi_bspline(x, ths->n[dim], ths->m, ths->spline_coeffs);
  else if(ths->pnfft_flags & PNFFT_WINDOW_SINC_POWER)
    return psi_sinc_power(x, ths->n[dim], ths->b[dim], ths->m);
  else if(ths->pnfft_flags & PNFFT_WINDOW_BESSEL_I0)
    return psi_bessel_i0(x, ths->n[dim], ths->b[dim], ths->m);
  else
    return psi_kaiser(x, ths->n[dim], ths->b[dim], ths->m);
}

R PNX(dpsi)(
    const PNX(plan) ths, int dim, R x
    )
{
  if(ths->pnfft_flags & PNFFT_WINDOW_GAUSSIAN)
    return dpsi_gaussian(x, ths->n[dim], ths->b[dim]);
  else if(ths->pnfft_flags & PNFFT_WINDOW_BSPLINE)
    return dpsi_bspline(x, ths->n[dim], ths->m, ths->spline_coeffs);
  else if(ths->pnfft_flags & PNFFT_WINDOW_SINC_POWER)
    return dpsi_sinc_power(x, ths->n[dim], ths->b[dim], ths->m);
  else if(ths->pnfft_flags & PNFFT_WINDOW_BESSEL_I0)
    return dpsi_bessel_i0(x, ths->n[dim], ths->b[dim], ths->m);
  else
    return dpsi_kaiser(x, ths->n[dim], ths->b[dim], ths->m);
}



static R psi_gaussian(
    R x, INT n, R b
    )
{
  return PNFFT_EXP( -PNFFT_SQR(n*x)/b ) / PNFFT_SQRT(PNFFT_PI*b);
}

static R dpsi_gaussian(
    R x, INT n, R b
    )
{
  return 2.0 * n/b * n*x * PNFFT_EXP( -PNFFT_SQR(n*x)/b ) / PNFFT_SQRT(PNFFT_PI*b);
}

static R psi_bspline(
    R x, INT n, int m, R *spline_coeffs
    )
{
  /* Bspline is shifted by m */
  return PNX(bspline)(2*m, n*x+m, spline_coeffs);
}

static R dpsi_bspline(
    R x, INT n, int m, R *spline_coeffs
    )
{
  /* Bspline is shifted by m */
  return (R)n *( PNX(bspline)(2*m-1, n*x + m - 1, spline_coeffs)
      - PNX(bspline)(2*m-1, n*x + m, spline_coeffs) );
}

static R psi_sinc_power(
    R x, INT n, R b, int m
    )
{
  return PNFFT_POW
    (
     PNX(sinc)( PNFFT_PI * (n*x) / b),
     K(2.0)*(R)m
    ) / b;
}

static R dpsi_sinc_power(
    R x, INT n, R b, int m
    )
{
  R y =  PNFFT_PI * n * x / b;
  if(PNFFT_FABS(y) > PNFFT_EPSILON)
    return 2.0 * (R)m * PNFFT_PI * (R)n / b * ( 1.0/y - 1.0/PNFFT_TAN(y) ) * psi_sinc_power(x, n, b, m);
  else
    return 0.0;
}

static R psi_bessel_i0(
    R x, INT n, R b, int m
    )
{
  return window_bessel_i0_1d(x, n, b, m);
}

static R dpsi_bessel_i0(
    R x, INT n, R b, int m
    )
{
  return window_bessel_i0_derivative_1d(x, n, b, m);
}

static R psi_kaiser(
    R x, INT n, R b, int m
    )
{
  return kaiser_bessel_1d(x, n, b, m);
}

static R dpsi_kaiser(
    R x, INT n, R b, int m
    )
{
  return kaiser_bessel_derivative_1d(
      x, n, b, m, kaiser_bessel_1d(x, n, b, m));
}

static R interpolation_odd(
    R x, R epsI, int order, INT num_nodes,
    const R *table
    )
{
  int sign = (x<0) ? -1 : 1;

  x = PNFFT_FABS(x);
  switch(order){
    case 1: return sign*intpol_odd_lin(x, table, num_nodes, epsI);
    case 2: return sign*intpol_odd_quad(x, table, num_nodes, epsI);
    default: return sign*intpol_odd_kub(x, table, num_nodes, epsI);
  }
}

static R interpolation_even(
    R x, R epsI, int order, INT num_nodes,
    const R *table
    )
{
  x = PNFFT_FABS(x);
  switch(order){
    case 1: return intpol_even_lin(x, table, num_nodes, epsI);
    case 2: return intpol_even_quad(x, table, num_nodes, epsI);
    default: return intpol_even_kub(x, table, num_nodes, epsI);
  }
}


/** linear spline interpolation in near field with even windows */
static R intpol_even_lin(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c3;
  int r;
  R f1,f2;

  c=x*num_nodes/eps_I;
  r=c; r=abs(r);
  f1=table[r];f2=table[r+1];
  c=fabs(c);
  c1=c-r;
  c3=c1-1.0;
  return (-f1*c3+f2*c1);
}

/** quadratic spline interpolation in near field with even windows */
static R intpol_even_quad(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c2,c3;
  int r;
  R f0,f1,f2;

  c=x*num_nodes/eps_I;
  r=c; r=abs(r);
  if (r==0) {f0=table[r+1];f1=table[r];f2=table[r+1];}
  else { f0=table[r-1];f1=table[r];f2=table[r+1];}
  c=fabs(c);
  c1=c-r;
  c2=c1+1.0;
  c3=c1-1.0;
  return (f0*c1*c3/2.0-f1*c2*c3+f2*c2*c1/2.0);
}

/** cubic spline interpolation in near field with even windows */
static R intpol_even_kub(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c2,c3,c4;
  int r;
  R f0,f1,f2,f3;

  c=x*num_nodes/eps_I;
  r=c; r=abs(r);
  if (r==0) {f0=table[r+1];f1=table[r];f2=table[r+1];f3=table[r+2];}
  else { f0=table[r-1];f1=table[r];f2=table[r+1];f3=table[r+2];}
  c=fabs(c);
  c1=c-r;
  c2=c1+1.0;
  c3=c1-1.0;
  c4=c1-2.0;
  return(-f0*c1*c3*c4/6.0+f1*c2*c3*c4/2.0-f2*c2*c1*c4/2.0+f3*c2*c1*c3/6.0);
}


/** linear spline interpolation in near field with odd windows */
static R intpol_odd_lin(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c3;
  int r, sign;
  R f1,f2;

  c=x*num_nodes/eps_I;
  sign = (c<0) ? -1 : 1;
  r=c; r=abs(r);
  f1=table[r];f2=table[r+1];
  c=fabs(c);
  c1=c-r;
  c3=c1-1.0;
  return sign*(-f1*c3+f2*c1);
}

/** quadratic spline interpolation in near field with odd windows */
static R intpol_odd_quad(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c2,c3;
  int r, sign;
  R f0,f1,f2;

  c=x*num_nodes/eps_I;
  sign = (c<0) ? -1 : 1;
  r=c; r=abs(r);
  if (r==0) {f0=-table[r+1];f1=table[r];f2=table[r+1];}
  else { f0=table[r-1];f1=table[r];f2=table[r+1];}
  c=fabs(c);
  c1=c-r;
  c2=c1+1.0;
  c3=c1-1.0;
  return sign*(f0*c1*c3/2.0-f1*c2*c3+f2*c2*c1/2.0);
}

/** cubic spline interpolation in near field with odd windows */
static R intpol_odd_kub(
    const R x, const R *table,
    const INT num_nodes, const R eps_I
    )
{
  R c,c1,c2,c3,c4;
  int r, sign;
  R f0,f1,f2,f3;

  c=x*num_nodes/eps_I;
  sign = (c<0) ? -1 : 1;
  r=c; r=abs(r);
  if (r==0) {f0=-table[r+1];f1=table[r];f2=table[r+1];f3=table[r+2];}
  else { f0=table[r-1];f1=table[r];f2=table[r+1];f3=table[r+2];}
  c=fabs(c);
  c1=c-r;
  c2=c1+1.0;
  c3=c1-1.0;
  c4=c1-2.0;
  return sign*(-f0*c1*c3*c4/6.0+f1*c2*c3*c4/2.0-f2*c2*c1*c4/2.0+f3*c2*c1*c3/6.0);
}



